#!/usr/bin/python
"""This scripts identifies the blamee when an autopkgtest fails
   and sends a notification
"""
# Copyright (C) 2013, Canonical Ltd (http://www.canonical.com/)
#
# Author: Jean-Baptiste Lallement <jean-baptiste.lallement@canonical.com>
#
# This software is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.
#
from launchpadlib.launchpad import Launchpad
import logging
import re
import urllib2
import argparse
from pprint import pprint
from os import path

LPDIR = path.expanduser('~/.lplib.adt')
LPCRED = path.join(LPDIR, 'adt.cred')

BLACKLIST = [
    'ps-jenkins@lists.canonical.com',
    'archive@ubuntu.com'
]


def set_logging(debugmode=False):
    """Initialize logging"""
    logging.basicConfig(
        level=logging.DEBUG if debugmode else logging.INFO,
        format="%(asctime)s %(levelname)s %(message)s")
    logging.debug('Debug mode enabled')


class AdtNotify(object):
    """This class identifies who should be notified after an autopkgtest
    failure and sends a notification to the blamee
    """
    def __init__(self, package=None, filename=None, release=None):
        """Constructor"""
        if package and '=' in package:
            (self.source_package, self.package_version) = \
                package.split('=', 1)
        else:
            (self.source_package, self.package_version) = \
                (package, None)
        self.filename = filename
        self.serie_name = release
        self.notifications = []

        logging.debug("Connecting to Launchpad")
        self.lauchpad = Launchpad.login_with('jibel', 'production',
                                             credentials_file=LPCRED,
                                             launchpadlib_dir=LPDIR)
        #self.lauchpad = Launchpad.login_anonymously('adt', 'production')
        self.distro = self.lauchpad.distributions['ubuntu']
        self.archive = self.distro.getArchive(name='primary')
        self.serie = self.distro.getSeries(name_or_version=self.serie_name)

    def collect(self):
        """Collect package and persons to notify from a list of source packages
        """
        if self.filename:
            with open(self.filename, 'r') as f:
                lines = f.readlines()
        else:
            lines = [("%s %s" % (self.source_package,
                                 "" if not self.package_version
                                 else self.package_version))]

        for line in lines:
            line = line.strip()
            if ' ' in line:
                (pkgname, pkgvers) = line.split(' ')
            else:
                (pkgname, pkgvers) = (line, None)
            logging.debug("= = = Processing %s %s = = =", pkgname, pkgvers)
            pub = self.__find_publication(pkgname, pkgvers)
            if pub is None:
                logging.warning("No publication found for package '%s'" %
                                pkgname)
                continue
            (blamee, blamee_email) = self.__find_person_to_notify(pub)

            version = pub.source_package_version
            pocket = pub.pocket

            if blamee is None or blamee_email is None:
                logging.warning("No blamee found for publication:  %s %s %s. "
                                "Skipping!" %
                                (pkgname, pocket, version))
                continue

            logging.info("Found publication: %s %s %s %s %s",
                         pkgname, pocket, version, blamee.name, blamee_email)

            if blamee_email is None:
                logging.warning("Notification email address not found for "
                                "blamee '%s'. Skipping!", blamee.name)
                continue
            if blamee_email in BLACKLIST:
                logging.warning("Blamee found but blacklisted. Skipping!")
                continue
            self.notifications.append((pkgname, version, pub.date_published,
                                       blamee.display_name, blamee_email))

    def send(self):
        """Send the notification
        """
        if self.notifications:
            logging.info("Sending %d notifications", len(self.notifications))
            pprint(self.notifications)

    def __find_publication(self, srcname, version=None):
        """Finds the publication of package and optional version starting
        searching into -proposed then searching in all the pockets if not found
        in -proposed
        """
        logging.debug("Searching publication for %s %s in -proposed",
                      srcname, "(latest)" if version is None else version)
        try:
            pub = self.archive.getPublishedSources(
                exact_match=True, source_name=srcname, pocket='Proposed',
                distro_series=self.serie, version=version)[0]
        except IndexError:
            pub = None
        if not pub:
            logging.debug("Publication not found in -proposed, "
                          "searching all pockets")
            try:
                pub = self.archive.getPublishedSources(
                    exact_match=True, source_name=srcname,
                    distro_series=self.serie, version=version)[0]
            except IndexError:
                return None
        return pub

    def __find_person_to_notify(self, pub):
        """Find the right person to notify about this upload"""
        blamee = None
        blamee_role = 'signer'
        if pub.package_signer:
            blamee = pub.package_signer
            logging.debug("Uploaded signed by '%s'" % blamee.name)
        elif pub.creator:
            logging.debug("No signer (sync request) using creator")
            blamee = pub.creator
            blamee_role = 'creator'

        emailaddr = ""
        # Automated copy, no changes file and no creator or signer records
        if blamee is None:
            return (None, None)
        try:
            emailaddr = blamee.preferred_email_address.email
        except ValueError:
            # Secret address search Changed-By in changelog
            logging.warning("No public email address for %s. "
                            "Falling back to changelog", blamee.name)
            (name, emailaddr) = self.__get_email_from_changes(pub, 'signer')
            if emailaddr is None:
                (name, emailaddr) = self.__get_email_from_changes(pub,
                                                                  'creator')
        return (blamee, emailaddr)

    def __get_email_from_changes(self, pub, role='signer'):
        """Get changed-by from changes file"""
        roles = {'signer': 'Signed-By',
                 'creator': 'Changed-By'}

        if role not in roles:
            logging.warning("Unrecognized role '%s'", role)
            return(None, None)
        cbre = re.compile("^%s: (?P<name>[^<]*) <(?P<email>[^>]*)>" %
                          roles[role])

        changesfileurl = pub.changesFileUrl()

        if changesfileurl is None:
            logging.warning("No changes files URL!")
            return(None, None)
        logging.debug("ChangesFile: %s", changesfileurl)
        changelog = urllib2.urlopen(changesfileurl)
        for line in changelog:
            cbm = cbre.match(line)
            if cbm:
                return (cbm.group('name'), cbm.group('email'))
        logging.warning("No address found for '%s' in changelog" % role)
        return(None, None)


def parse_args():
    """Parse command line arguments
    """
    try:
        from distro_info import UbuntuDistroInfo
        default_release = UbuntuDistroInfo().devel()
    except ImportError:
        default_release = "saucy"
        logging.warning("Failed to import distro_info! default release set "
                        "to '%s'", default_release)

    parser = argparse.ArgumentParser(
        description='Report results of autopkgtest')
    source = parser.add_mutually_exclusive_group(required=True)
    parser.add_argument('-d', '--debug', action='store_true',
                        default=False, help='enable debug mode')
    source.add_argument('-p', '--package', metavar='PACKAGE',
                        help="Source package name with format "
                        "srcpkg[=version]")
    source.add_argument('-i', '--input-file', metavar='INPUTFILE',
                        help="Read packages and versions from a file. The "
                        "file contains 1 line per package/version. "
                        "Package and version must be separated by a "
                        "space.")
    parser.add_argument('-r', '--release', metavar='RELEASE',
                        default=default_release,
                        help='Release (default=%s)' % default_release)

    args = parser.parse_args()
    set_logging(args.debug)

    return args

if __name__ == "__main__":
    ARGS = parse_args()

    ADT_NOTIFY = AdtNotify(ARGS.package, ARGS.input_file, ARGS.release)
    ADT_NOTIFY.collect()
    notifications = ADT_NOTIFY.notifications
    for notification in notifications:
        print("%s %s uploaded on %s by %s <%s>\n" % notification)
