#! /usr/bin/python
""" Interface between britney and autopkgtest running on Jenkins

Apt Cache management class
"""
# Copyright (C) 2012, Canonical Ltd (http://www.canonical.com/)
#
# Author: Jean-Baptiste Lallement <jean-baptiste.lallement@canonical.com>
#
# This software is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.
#
import apt
import time
import os
import logging
from urlparse import urlparse


class AptCache(apt.Cache):
    '''subclass of apt.Cache that handles cache updates'''
    supported_archs = ('amd64', 'i386')

    def __init__(self, rootdir, cache_age=10):
        '''Update apt cache

        :param rootdir: Alternative apt root directory
        :param cache_age: Update the cache when it is older than cache_age
        :param force_update: Force cache update even if cache_age
            is not reached

        '''
        logging.debug("Instantiating new cache using rootdir '%s'", rootdir)
        self.cache_age = cache_age
        self.arch = None

        try:
            import distro_info
            self.release = distro_info.UbuntuDistroInfo().devel()
        except:
            self.release = 'trusty'
        self.uri = None
        self.components = ['main']
        self.proposed = False

        self.timestampfile = os.path.normpath(
            os.path.join(rootdir, 'var/cache/apt/pkgcache.bin'))

        super(AptCache, self).__init__(rootdir=rootdir, memonly=False)
        self.rootdir = rootdir

    def update(self, cache_age=60, force_update=False):
        ''' Equivalent of apt-get update

        :param cache_age: Update the cache when it is older than
            cache_age (minutes)
        :param force_update: Force cache update even if cache_age
            is not reached
        :return: True on success
        '''
        currtime = time.time()
        timestamp = 0

        if os.path.exists(self.timestampfile):
            timestamp = os.path.getmtime(self.timestampfile)

        if force_update or (
            not os.path.exists(self.timestampfile)) or (
                (currtime - timestamp) > cache_age * 60):
            logging.debug('Updating cache')
            super(AptCache, self).update()
            super(AptCache, self).open()
        else:
            logging.debug('Not updating. Timestamp age: %d min',
                          (currtime - timestamp) / 60)

    def write_sources_list(self, uri, arch=None, release=None,
                           components=None, proposed=False,
                           overwrite_sources=False):
        """ Write sources.list file """
        logging.debug('Writing sources.list')

        if not uri:
            logging.error('No \'apturi\' defined!')
            return False

        self.uri = uri
        if arch is not None:
            if not arch in self.supported_archs:
                logging.error('Unsupported architecture: %s', arch)
                return False
            self.arch = arch
        if release is not None:
            self.release = release
        if components is not None:
            if type(components) == str:
                self.components = [components]
            else:
                self.components = components
        self.proposed = proposed

        parsed = urlparse(self.uri)
        if parsed.scheme == 'file':
            self.uri = parsed.scheme + '://' + os.path.abspath(
                os.path.expanduser(parsed.path))

        archbit = ""
        if self.arch is not None:
            archbit = "[ arch={} ]".format(self.arch)

        src = """\
deb {arch} {uri} {release} {components}
deb-src {uri} {release} {components}
""".format(arch=archbit, uri=self.uri, release=self.release,
           components=' '.join(self.components))

        if self.proposed:
            src += """\
deb {arch} {uri} {release}-proposed {components}
deb-src {uri} {release}-proposed {components}
""".format(arch=archbit, uri=self.uri, release=self.release,
           components=' '.join(self.components))

        # PPA support disabled for the moment
        #if self.args.ppas:
        #    for ppa in self.args.ppas:
        #        (ppa_line, ignored) = expand_ppa_line(ppa, release)
        #        ppa_src_line = 'deb-src' + ppa_line[3:]
        #        src += "\n\n%s\n%s\n"% (ppa_line, ppa_src_line)
        path = os.path.join(self.rootdir, 'etc/apt/sources.list')
        logging.debug('sources.list to %s', path)
        if not os.path.exists(path):
            os.makedirs(os.path.dirname(path))
        else:
            if not overwrite_sources and os.path.getsize(path) > 0:
                logging.warning('sources.list exists and overwrite set to '
                                'false, not overwriting!')
                return
        with open(path, 'w') as f:
            f.write(src)

        self.update(force_update=True)
        return True
