#! /usr/bin/python
#
# Test Cache functions
#

# Copyright (C) 2013 Canonical
#
# Authors: Jean-Baptiste Lallement <jean-baptiste.lallement@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

import unittest
from testscenarios import TestWithScenarios
import apt
from aptcache import AptCache
from tempfile import mkdtemp
import os
import shutil
from distro_info import UbuntuDistroInfo
import mock
import re

URI_OFFICIAL_MIRROR = "http://archive.ubuntu.com/ubuntu/"
URI_FILE = os.path.join("file:///tmp/", __name__, "ubuntu")
DEFAULT_RELEASE_DEV = UbuntuDistroInfo().devel()
DEFAULT_RELEASE_LTS = UbuntuDistroInfo().lts()
ARCH_NATIVE = "amd64"
ARCH_NONNATIVE = "i386"
ARCH_INVALID = "jibel512"


def sources_list_is_valid(aptroot, uri, arch=None, release=None,
                          components=None, proposed=False):
    """ Verify that sources.list in aptroot is well-formed and matches the
    arguments
    """
    if not release:
        release = DEFAULT_RELEASE_DEV
    if not components:
        components = "main"

    sourceslist = os.path.join(aptroot, "etc", "apt", "sources.list")
    if not os.path.isfile(sourceslist):
        return False

    with open(sourceslist, 'r') as fhd:
        content = fhd.readlines()

    re_entry = re.compile("^(?P<deb>deb|deb-src)\s+(?P<arch>\[\s*arch=\w+\s*\])?\s*(?P<uri>\S+)\s+(?P<distro>\S+)\s+(?P<components>.*)")
    entries = []
    for entry in content:
        mre = re_entry.match(entry.rstrip())
        if mre:
            try:
                entries.append({
                    "deb": mre.group("deb"),
                    "arch": mre.group("arch"),
                    "uri": mre.group("uri"),
                    "distro": mre.group("distro"),
                    "components": mre.group("components").split(),
                })
            except IndexError:
                # IndexError means an expected component is not present in the
                # sources list entry
                return False

    # Check that binary and source entries are present
    debs = [entry['deb'] for entry in entries]
    if ((not proposed and debs != ['deb', 'deb-src']) or
            (proposed and debs != ['deb', 'deb-src'] * 2)):
        return False

    # If an arch is specified only binary entries must have arch
    if arch:
        for entry in entries:
            if entry['deb'] == "deb":
                if not arch in entry['arch']:
                    return False
            elif entry['deb'] == "deb-src":
                if entry['arch'] is not None:
                    return False

    # Check that uri is correct
    entry_uri = set([entry['uri'] for entry in entries])
    if len(entry_uri) != 1 or uri not in entry_uri:
        return False

    entry_distro = set([entry['distro'] for entry in entries])
    if ((not proposed and len(entry_distro) != 1) or
            (proposed and len(entry_distro) != 2) or
            (release not in entry_distro) or
            (proposed and release + "-proposed" not in entry_distro)):
        return False

    entry_components = set()
    for entry in entries:
        entry_components |= set(entry['components'])
    if type(components) == str:
        if components not in entry_components:
            return False
    elif type(components) == list:
        for comp in components:
            if comp not in entry_components:
                return False
    elif components is not None:
        # Other type are supported
        return False

    return True


class TestCacheSourcesList(TestWithScenarios, unittest.TestCase):
    """ Test the functionality of the apt cache """
    scenarios = [
        ("uri_official_mirror",
         {"uri": URI_OFFICIAL_MIRROR,
          "arch": None,
          "release": None,
          "components": None,
          "proposed": False,
          "return_code": True}),
        ("uri_empty",
         {"uri": None,
          "arch": None,
          "release": None,
          "components": None,
          "proposed": False,
          "return_code": False}),
        ("uri_file",
         {"uri": URI_FILE,
          "arch": None,
          "release": None,
          "components": None,
          "proposed": False,
          "return_code": True}),
        ("release_lts",
         {"uri": URI_OFFICIAL_MIRROR,
          "arch": None,
          "release": DEFAULT_RELEASE_LTS,
          "components": None,
          "proposed": False,
          "return_code": True}),
        ("arch_invalid",
         {"uri": URI_OFFICIAL_MIRROR,
          "arch": ARCH_INVALID,
          "release": None,
          "components": None,
          "proposed": False,
          "return_code": False}),
        ("arch_nonnative",
         {"uri": URI_OFFICIAL_MIRROR,
          "arch": ARCH_NONNATIVE,
          "release": None,
          "components": None,
          "proposed": False,
          "return_code": True}),
        ("arch_nonnative_with_proposed",
         {"uri": URI_OFFICIAL_MIRROR,
          "arch": ARCH_NONNATIVE,
          "release": None,
          "components": None,
          "proposed": True,
          "return_code": True}),
        ("components_single",
         {"uri": URI_OFFICIAL_MIRROR,
          "arch": None,
          "release": None,
          "components": "universe",
          "proposed": False,
          "return_code": True}),
        ("components_multiple",
         {"uri": URI_OFFICIAL_MIRROR,
          "arch": None,
          "release": None,
          "components": ["main", "universe", "restricted"],
          "proposed": False,
          "return_code": True}),
        ("proposed_enabled",
         {"uri": URI_OFFICIAL_MIRROR,
          "arch": None,
          "release": None,
          "components": None,
          "proposed": True,
          "return_code": True}),
    ]

    def setUp(self):
        self.native_arch = apt.apt_pkg.config.find("APT::Architecture")
        apt.apt_pkg.config.set("APT::Architecture", ARCH_NATIVE)
        self.addCleanup(lambda: apt.apt_pkg.config.set("APT::Architecture", self.native_arch))
        self.aptroot = mkdtemp(prefix=__name__ + ".")
        apt.apt_pkg.config.set('Dir', self.aptroot)
        self.cache = AptCache(rootdir=self.aptroot)
        self.aptcache_update = apt.Cache.update
        self.aptcache_open = apt.Cache.open
        apt.Cache.update = mock.Mock()
        apt.Cache.open = mock.Mock()

    def test_sourcelist(self):
        """ Write a valid a sources.list

        Test if arguments passed to write_sources_list are valid and if the
        resulting sources.list file is valid
        """
        retcode = self.cache.write_sources_list(
            self.uri,
            self.arch,
            self.release,
            self.components,
            self.proposed,
            overwrite_sources=True)

        # By default source list should contain entries for main on the
        # development release
        self.assertEqual(self.return_code, retcode,
                         "Return code doesn't match expected result")

        # Additional verifications if expected return code is True
        if self.return_code:
            self.assertTrue(
                sources_list_is_valid(
                    self.aptroot,
                    self.uri,
                    self.arch,
                    self.release,
                    self.components,
                    self.proposed))

            self.assertTrue(apt.Cache.update.called, "Cache has not been updated")
            self.assertTrue(apt.Cache.open.called, "Cache has not been opened")

    def tearDown(self):
        apt.Cache.update = self.aptcache_update
        apt.Cache.open = self.aptcache_open
        if os.path.isdir(self.aptroot):
            shutil.rmtree(self.aptroot)


class TestCacheMisc(unittest.TestCase):
    """ Test the functionality of the apt cache """
    def setUp(self):
        self.native_arch = apt.apt_pkg.config.find("APT::Architecture")
        apt.apt_pkg.config.set("APT::Architecture", "amd64")
        self.addCleanup(lambda: apt.apt_pkg.config.set("APT::Architecture", self.native_arch))
        self.aptroot = mkdtemp(prefix=__name__ + ".")
        apt.apt_pkg.config.set('Dir', self.aptroot)
        self.cache = AptCache(rootdir=self.aptroot)

        self.aptcache_update = apt.Cache.update
        self.aptcache_open = apt.Cache.open
        apt.Cache.update = mock.Mock()
        apt.Cache.open = mock.Mock()

    def test_cache_created_empty(self):
        """ Test integrity of the apt directory structure """
        empty_tree = set([
            'etc',
            'etc/apt',
            'etc/apt/sources.list',
            'var',
            'var/cache',
            'var/cache/apt',
            'var/cache/apt/archives',
            'var/cache/apt/archives/partial',
            'var/cache/apt/pkgcache.bin',
            'var/cache/apt/srcpkgcache.bin',
            'var/lib',
            'var/lib/apt',
            'var/lib/apt/lists',
            'var/lib/apt/lists/partial',
            'var/lib/dpkg',
            'var/lib/dpkg/status'
        ])

        dirtree = []
        for root, dirs, files in os.walk(self.aptroot):
            basedir = root.replace(self.aptroot, '').strip('/')
            dirtree.extend([os.path.join(basedir, dir) for dir in dirs + files])

        created_tree = set(dirtree)
        self.assertEqual(empty_tree, created_tree)

    def test_cache_update_noforce(self):
        """Test cache update"""
        self.cache.update()
        self.assertTrue(apt.Cache.update.called, "Cache has not been updated")
        self.assertTrue(apt.Cache.open.called, "Cache has not been opened")

    def test_cache_update_force(self):
        """Test cache update"""
        self.cache.update(force_update=True)
        self.assertTrue(apt.Cache.update.called, "Cache has not been updated")
        self.assertTrue(apt.Cache.open.called, "Cache has not been opened")

    def test_sourcelist_no_overwrite_nonempty(self):
        """ Check that existing sources.list is not overwritten

        """
        apt.Cache.update = mock.Mock()
        sourceslist = os.path.join(self.aptroot, "etc", "apt", "sources.list")
        self.cache.write_sources_list(URI_FILE)
        stat_before = os.stat(sourceslist)
        self.cache.write_sources_list(URI_OFFICIAL_MIRROR)
        stat_after = os.stat(sourceslist)
        self.assertEqual(stat_before, stat_after)
        #self.assertFalse(apt.Cache.update.called, "Cache has been updated")

    def test_sourcelist_overwrite_empty(self):
        """ Check that an existing empty sources.list is overwritten

        """
        apt.Cache.update = mock.Mock()
        sourceslist = os.path.join(self.aptroot, "etc", "apt", "sources.list")
        with open(sourceslist, 'w+'): pass
        stat_before = os.stat(sourceslist)
        self.cache.write_sources_list(URI_OFFICIAL_MIRROR)
        stat_after = os.stat(sourceslist)
        self.assertNotEqual(stat_before, stat_after)
        self.assertTrue(apt.Cache.update.called, "Cache has not been updated")

    def tearDown(self):
        apt.Cache.update = self.aptcache_update
        apt.Cache.open = self.aptcache_open
        if os.path.isdir(self.aptroot):
            shutil.rmtree(self.aptroot)

if __name__ == "__main__":
    unittest.main()
